#include "stdafx.h"
#include <math.h>
#include "Component.h"
#include "MainTaskDlg.h"
#include "RMA.h"
CCompNodeList gCompNodeList;
CTaskNodeList gTaskNodeList;
CProcessorNodeList gProcessorNodeList;
//#define DEBUG
//Assume only one interaction model for now, since for multiple interaction models
//we need to consider component references between interaction models.
//Assume unique names for each component, since I use FindCompNodeByName()

bool CheckUniqueNames();
CCompNode *FindCompNodeByName(const CString &name);
void ProcessTimer(CBuilderAtom *timer);
void ProcessComp(CBuilderModel *comp);
void ProcessPublishBase(CBuilderAtom *pubBase, CCompNode *rootCompNode);
void ProcessReceptacle(const CBuilderObject *receptacle, CCompNode *rootCompNode);
void CheckForCyclesAndAssignRates(const CBuilderAtomList *timers);
void DFSVisit(CCompNode *u, int timerPeriod);
void PrintCycle(CCompNode *u, CCompNode *v);
void ConstructTaskGraph();

CProcessorNode *GetProcessorNodeByID(int id);
void SortTasksByPriority(CTaskNodeList & taskNodeList);
void QuickSort(CTaskNode **data, int lower, int upper);


void DebugDisplayProcessorNodeList();
void DebugDisplayTaskNodeList();
void DebugDisplayCompNodeList();

void CComponent::InvokeEx(CBuilder &builder,CBuilderObject *focus, CBuilderObjectList &selected, long param) 
{	
//Testing Code ---------------------------------------------------
/*	int counter = 1;
	CTaskNodeList tasks;
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 7));
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 4));
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 19));
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 3));
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 8));
	tasks.AddTail(new CTaskNode(counter++, "", 1, 1, 7));
    SortTasksByPriority(tasks);
	POSITION poss = tasks.GetHeadPosition();
	while(poss) {
		CTaskNode *taskNode = tasks.GetNext(poss);
		CString cs;
		cs.Format("%d:%d", taskNode->id, taskNode->priority);
		AfxMessageBox(cs);
	}
	return;
*/
//End of testing code --------------------------------------------



	gCompNodeList.RemoveAll();
	gTaskNodeList.RemoveAll();
	gProcessorNodeList.RemoveAll();
	int taskCounter = 1, compCounter = 1;


	//Construct gProcessorNodeList and gTaskNodeList
	const CBuilderFolder *configFolder = builder.GetFolder("Configurations");
	ASSERT(configFolder);
	const CBuilderModelList *configs = configFolder->GetRootModels();
	if(!configs) {
		AfxMessageBox("No configs are defined!");
		return;
	}

	POSITION pos = configs->GetHeadPosition();
	while(pos) {
		CBuilderModel *config = configs->GetNext(pos);
		const CBuilderModelList *processors = config->GetModels();
		POSITION pos2 = processors->GetHeadPosition();
		while(pos2) {
			CBuilderModel *processor = processors->GetNext(pos2);
			int processorID;
			CString processorName;
			double cxtSwitchTime;
			processor->GetAttribute("ProcessorID", processorID);
			processorName = processor->GetName();
			processor->GetAttribute("CxtSwitchTime", cxtSwitchTime);
			CProcessorNode *processorNode = new CProcessorNode(processorID, processorName, cxtSwitchTime);
			gProcessorNodeList.AddTail(processorNode);

			const CBuilderAtomList *threads = processor->GetAtoms("Thread");
			POSITION pos3 = threads->GetHeadPosition();
			while(pos3) {
				CBuilderAtom *thread = threads->GetNext(pos3);
				int threadPeriod, threadPriority;
				thread->GetAttribute("ThreadPeriod", threadPeriod);
				thread->GetAttribute("ThreadPriority", threadPriority);
				CString temp;
				temp.Format("%s_Period_%d", processorName, threadPeriod);
				CTaskNode *taskNode = new CTaskNode(taskCounter++, temp, processorID, threadPeriod, threadPriority);
				gTaskNodeList.AddTail(taskNode);
				processorNode->taskNodeList.AddTail(taskNode);
			}
			SortTasksByPriority(processorNode->taskNodeList);
		}
	}


	const CBuilderFolder *interactionFolder = builder.GetFolder("Interactions"); 
    ASSERT(interactionFolder);
	const CBuilderModelList *rootModels = interactionFolder->GetRootModels();
	ASSERT(rootModels);

	
	pos = rootModels->GetHeadPosition(); 
    while(pos) { 
        CBuilderModel *rootModel = rootModels->GetNext(pos); 

		//Grab all Timers and put them in the global list. Treat timer just like a regular
		//component.
		const CBuilderAtomList *timers = rootModel->GetAtoms("Timer");
		POSITION pos = timers->GetHeadPosition();
		while(pos) {
			CBuilderAtom *timer = timers->GetNext(pos);
			int timerPeriod;
			timer->GetAttribute("Period", timerPeriod);
			int processorID;
			timer->GetAttribute("RunsOnProcessorID", processorID);
			CString cs;
			cs.Format("Timer_%d", timerPeriod);
			CCompNode *compNode = new CCompNode(compCounter++, cs, processorID, 0);
			compNode->primaryPeriod = timerPeriod;
			compNode->isTimer = true;
			gCompNodeList.AddTail(compNode);
		}

		//Grab all ComponentType and put them in the global list.
       	const CBuilderModelList *comps = rootModel->GetModels("ComponentType");
		POSITION pos2 = comps->GetHeadPosition();
		while(pos2) {
			CBuilderModel *comp = comps->GetNext(pos2);
			double wcet;
			comp->GetAttribute("CompWCET", wcet);
			int processorID;
			comp->GetAttribute("RunsOnProcessorID", processorID);
			CCompNode *compNode = new CCompNode(compCounter++, comp->GetName(), processorID, wcet);
			gCompNodeList.AddTail(compNode);
		}

		//Construct EDG for timers part.
		POSITION pos3 = timers->GetHeadPosition();
		while(pos3) {
			CBuilderAtom *timer = timers->GetNext(pos3);
			ProcessTimer(timer);
		}

		//Construct EDG for regular components part.
		POSITION pos4 = comps->GetHeadPosition();
		while(pos4) {
			CBuilderModel *comp = comps->GetNext(pos4);
			ProcessComp(comp);
		}

		if(!CheckUniqueNames()) {
			AfxMessageBox("Each processor, task and component must have a unique name for this version of software to work. Please edit your model.");
			return;
		}
		CheckForCyclesAndAssignRates(timers);

		ConstructTaskGraph();
		ComputeUtilizations();

		if(HolisticAnalysis()) {
			//AfxMessageBox("RMA Analysis converged!");
		} else {
			AfxMessageBox("RMA analysis did not converge!");
		}

		CMainTaskDlg dlg;
		dlg.DoModal();
#ifdef DEBUG
//		DebugDisplayProcessorNodeList();
//		DebugDisplayTaskNodeList();
//		DebugDisplayCompNodeList();
#endif
	}
}

bool CheckUniqueNames() {
	POSITION pos = gProcessorNodeList.GetHeadPosition();
	while(pos) {
		CProcessorNode *pNode = gProcessorNodeList.GetNext(pos);
		POSITION pos2 = gProcessorNodeList.GetHeadPosition();
		while(pos2) {
			CProcessorNode *pNode2 = gProcessorNodeList.GetNext(pos2);
			if(pNode->id != pNode2->id && pNode->name == pNode2->name) {
				CString cs;
				cs.Format("Duplicate processor names detected: %s. ", pNode->name);
				AfxMessageBox(cs);
				return false;
			}
		}
	}

	pos = gTaskNodeList.GetHeadPosition();
	while(pos) {
		CTaskNode *pNode = gTaskNodeList.GetNext(pos);
		POSITION pos2 = gTaskNodeList.GetHeadPosition();
		while(pos2) {
			CTaskNode *pNode2 = gTaskNodeList.GetNext(pos2);
			if(pNode->id != pNode2->id && pNode->name == pNode2->name) {
				CString cs;
				cs.Format("Duplicate task names detected: %s. ", pNode->name);
				AfxMessageBox(cs);
				return false;
			}
		}
	}

	pos = gCompNodeList.GetHeadPosition();
	while(pos) {
		CCompNode *pNode = gCompNodeList.GetNext(pos);
		POSITION pos2 = gCompNodeList.GetHeadPosition();
		while(pos2) {
			CCompNode *pNode2 = gCompNodeList.GetNext(pos2);
			if(pNode->id != pNode2->id && pNode->name == pNode2->name) {
				CString cs;
				cs.Format("Duplicate component names detected: %s. ", pNode->name);
				AfxMessageBox(cs);
				return false;
			}
		}
	}

	return true;
}

void SortTasksByPriority(CTaskNodeList & taskNodeList) {
	int numTasks = taskNodeList.GetCount();
	CTaskNode **taskNodeArray = new CTaskNode *[numTasks];
	int i = 0;
	POSITION pos = taskNodeList.GetHeadPosition();
	while(pos) {
		taskNodeArray[i++] = taskNodeList.GetNext(pos);
	}

	QuickSort(taskNodeArray, 0, numTasks - 1);
	
	taskNodeList.RemoveAll();
	for(i = 0; i < numTasks; i++) {
		/*CString cs;
		cs.Format("SHit %d:%d", taskNodeArray[i]->id, taskNodeArray[i]->priority);
		AfxMessageBox(cs);*/
		taskNodeList.AddTail(taskNodeArray[i]);
	}
}

void swap(CTaskNode **data, int indexA, int indexB)
{
	CTaskNode *temp;
	temp = data[indexA];
	data[indexA] = data[indexB];
	data[indexB] = temp;
}

//Sort tasknodes in descending priority order.
void QuickSort(CTaskNode **data, int lower, int upper)
{
    if(lower>=upper)
        return;
    swap(data, lower, (lower+upper)/2);
    int m = lower;
    for(int i=lower+1; i<=upper; i++)
    {
        //if(strcmp(data[i], data[lower]) < 1)
        if(data[i]->priority > data[lower]->priority)  
			swap(data, ++m, i);
    }
    swap(data, lower, m);
    QuickSort(data, lower, m-1);
    QuickSort(data, m+1, upper);
}

void ConstructTaskGraph() {
	//Assign components to tasks based on component primaryPeriod and processorID.
	POSITION pos = gTaskNodeList.GetHeadPosition();
	while(pos) {
		CTaskNode *taskNode = gTaskNodeList.GetNext(pos);

		POSITION pos2 = gCompNodeList.GetHeadPosition();
		while(pos2) {
			CCompNode *compNode = gCompNodeList.GetNext(pos2);
			if(compNode->processorID == taskNode->processorID &&
				compNode->primaryPeriod == taskNode->period) {
				if(!compNode->isTimer) {
					taskNode->compNodeList.AddTail(compNode);
					taskNode->wcet += compNode->wcet;
				} else {
					taskNode->timerTriggered = true;
				}
			}
		}
	}

	//Construct task graph, precedence relationships are determined by component
	//graph precedence relationships.
	pos = gTaskNodeList.GetHeadPosition();
	while(pos) {
		CTaskNode *taskNode1 = gTaskNodeList.GetNext(pos);

		POSITION pos2 = gTaskNodeList.GetHeadPosition();
		while(pos2) {
			CTaskNode *taskNode2 = gTaskNodeList.GetNext(pos2);
			if(taskNode1 == taskNode2) continue;
			POSITION pos3 = taskNode1->compNodeList.GetHeadPosition();
			while(pos3) {
				CCompNode *compNode1 = taskNode1->compNodeList.GetNext(pos3);
				POSITION pos4 = compNode1->adjList.GetHeadPosition();
				while(pos4) {
					CCompNode *adjCompNode1 = compNode1->adjList.GetNext(pos4);
					POSITION pos5 = taskNode2->compNodeList.GetHeadPosition();
					while(pos5) {
						CCompNode *compNode2 = taskNode2->compNodeList.GetNext(pos5);
						//if(adjCompNode1 == compNode2) {
						if(adjCompNode1->name == compNode2->name) {
							taskNode1->adjList.AddTail(taskNode2);
						}
					}
				}
			}
		}
	}
}


void CheckForCyclesAndAssignRates(const CBuilderAtomList *timers) {
	POSITION pos = gCompNodeList.GetHeadPosition();
	while(pos) {
		CCompNode *u = gCompNodeList.GetNext(pos);
		u->color = WHITE;
		u->predecessor = NULL;
	}
/*
	pos = gCompNodeList.GetHeadPosition();
	while(pos) {
		CCompNode *u = gCompNodeList.GetNext(pos);
		if(u->color == WHITE) {
			DFSVisit(u);
		}
	}
*/
	if(!timers) {
		AfxMessageBox("No timers are defined!");
		return;
	}
	POSITION pos2 = timers->GetHeadPosition();
	while(pos2) {
		CBuilderAtom *timer = timers->GetNext(pos2);
		int timerPeriod;
		timer->GetAttribute("Period", timerPeriod);
		CString cs;
		cs.Format("Timer_%d", timerPeriod);
		CCompNode *rootCompNode = FindCompNodeByName(cs);
		ASSERT(rootCompNode);
		ASSERT(rootCompNode->color == WHITE);
		DFSVisit(rootCompNode, timerPeriod);
	}

	POSITION pos3 = gCompNodeList.GetHeadPosition();
	while(pos3) {
		CCompNode *u = gCompNodeList.GetNext(pos3);
		//Why is nodes color gray instead of black? maybe a bug in DFS code.
		/*CString cs;
		if(u->color == WHITE)
			cs.Format("Comp: %s Color WHITE", u->name);
		else if(u->color = GRAY)
			cs.Format("Comp: %s Color GRAY", u->name);
		else if(u->color = BLACK)
			cs.Format("Comp: %s Color BLACK", u->name);
		else ASSERT(0);
		AfxMessageBox(cs);
		*/
		if(u->color == WHITE) {
			CString cs;
			cs.Format("Component %s is not reached from any timer!", u->name);
			AfxMessageBox(cs);
		}
	}
}

void DFSVisit(CCompNode *u, int timerPeriod) {
	u->color = GRAY;
	POSITION pos = u->adjList.GetHeadPosition();
	while(pos) {
		CCompNode *v = u->adjList.GetNext(pos);
		if(v->color == GRAY) {
			PrintCycle(u, v);
		} else if(v->color == WHITE) {
			v->predecessor = u;
			v->primaryPeriod = timerPeriod;
			DFSVisit(v, timerPeriod);
		}
	}
	u->color = BLACK;
}


void PrintCycle(CCompNode *u, CCompNode *v) {
	CCompNode *u1 = u;
	CString cs(u1->name);
	while(u1 != v) {
		u1 = u1->predecessor;
		ASSERT(u1);
		cs = u1->name + "->" + cs;
	}
	cs = "EDG Cycle detected: " + cs;
	AfxMessageBox(cs);
}

/*
void AssignRates(CBuilderAtom *timer) {
	int timerPeriod;
	timer->GetAttribute("Period", timerPeriod);
	CString cs;
	cs.Format("Timer_%d", timerPeriod);

	CCompNode *rootCompNode = FindCompNodeByName(cs);
	ASSERT(rootCompNode);
	CRootCompNode *rootCompNode = FindRootCompNodeByName(initCompNode->name);
	ASSERT(rootCompNode);
	CCompNodeList *compNodeList = rootCompNode->compNodeList;
	POSITION pos2 = compNodeList->GetHeadPosition();
	while(pos2) {
		CCompNode *compNode = compNodeList->GetNext(pos2);
		if(compNode->primaryPeriod == -1) {
			compNode->primaryPeriod = timerPeriod;
		} else if(compNode->secondaryPeriod == -1) {
			compNode->secondaryPeriod = timerPeriod;
		} else {
			cs.Format("Error: Currently a maximum of 2 rates per component is supported, while %s has more than 2 rates. ",
				compNode->name);
			AfxMessageBox(cs);
		}
		AssignRates(compNode, timerPeriod);
	}
}

*/

CCompNode *FindCompNodeByName(const CString &name) {
	POSITION pos = gCompNodeList.GetHeadPosition();
	while(pos) {
		CCompNode *compNode = gCompNodeList.GetNext(pos);
		if(compNode->name == name) {
			return compNode;
		}
	}
	return NULL;
}

void ProcessTimer(CBuilderAtom *timer) {
	int timerPeriod;
	timer->GetAttribute("Period", timerPeriod);
	CString cs;
	cs.Format("Timer_%d", timerPeriod);
	CCompNode *rootCompNode = FindCompNodeByName(cs);
	ASSERT(rootCompNode);

	ProcessPublishBase(timer, rootCompNode);
}

void ProcessComp(CBuilderModel *comp) {
	CCompNode *rootCompNode = FindCompNodeByName(comp->GetName());
	ASSERT(rootCompNode);

	const CBuilderAtomList *pubPorts = comp->GetAtoms("PublishPort");
	if(pubPorts) {
		POSITION pos = pubPorts->GetHeadPosition();
		while(pos) {
			CBuilderAtom *pubPort = pubPorts->GetNext(pos);
			ProcessPublishBase(pubPort, rootCompNode);
		}
	}

	const CBuilderModelReferenceList *receptacles = comp->GetModelReferences("Receptacle");
	if(receptacles){
		POSITION pos = receptacles->GetHeadPosition();
		while(pos) {
			CBuilderModelReference *receptacleRef = receptacles->GetNext(pos);
			const CBuilderObject *receptacle = receptacleRef->GetReferred();
			ProcessReceptacle(receptacleRef, rootCompNode);
		}
	}
}
	

void ProcessPublishBase(CBuilderAtom *pubBase, CCompNode *rootCompNode) {
	CBuilderObjectList events;
	pubBase->GetOutConnectedObjects("Publish", events);
	POSITION pos2 = events.GetHeadPosition();
	while(pos2) {
		CBuilderObject *event = events.GetNext(pos2);
		CBuilderObjectList subPorts;
		event->GetOutConnectedObjects("Subscribe", subPorts);
		POSITION pos3 = subPorts.GetHeadPosition();
		//if(!pos3) {
		if(subPorts.GetCount() == 0) {
			CString cs;
			cs.Format("Error: no subscribers for the event %s published by the component %s. ",
				event->GetName(), pubBase->GetParent()->GetName());
			AfxMessageBox(cs);
			continue;
		}
		while(pos3) {
			CBuilderObject *subPort = subPorts.GetNext(pos3);
			const CBuilderModel *destComp = subPort->GetParent();
			CCompNode *compNode = FindCompNodeByName(destComp->GetName());
			ASSERT(compNode);
			rootCompNode->adjList.AddTail(compNode);
		}
	}
}

void ProcessReceptacle(const CBuilderObject *receptacle, CCompNode *rootCompNode) {
	const CBuilderConnectionList *invokes = receptacle->GetOutConnections("Invoke");
	POSITION pos2 = invokes->GetHeadPosition();
	//CBuilderObjectList invokes; 
	//((CBuilderObject *)receptacle)->GetDirectOutConnections("Invoke", invokes);
	while(pos2) {
		CBuilderObject *invoke = invokes->GetNext(pos2);
		CString methodName;
		invoke->GetAttribute("MethodName", methodName);
		if(methodName == "SetData") {
			CBuilderObject *facet = ((CBuilderConnection *)invoke)->GetDestination();
			const CBuilderModel *destComp = facet->GetParent();
			CCompNode *compNode = FindCompNodeByName(destComp->GetName());
			ASSERT(compNode);
			rootCompNode->adjList.AddTail(compNode);
		}
	}
}	

CProcessorNode *GetProcessorNodeByID(int id) {
	POSITION pos = gProcessorNodeList.GetHeadPosition();
	while(pos) {
		CProcessorNode *processorNode = gProcessorNodeList.GetNext(pos);
		if(id == processorNode->id) {
			return processorNode;
		}
	}
	return NULL;
}



void DebugDisplayProcessorNodeList() {
	POSITION pos = gProcessorNodeList.GetHeadPosition();
	while(pos) {
		CProcessorNode *processorNode = gProcessorNodeList.GetNext(pos);
		CString temp;
		temp.Format("Processor %d, name: %s", processorNode->id, processorNode->name);
		AfxMessageBox(temp);
	}
}

void DebugDisplayTaskNodeList() {
	POSITION pos = gTaskNodeList.GetHeadPosition();
	while(pos) {
		CTaskNode *taskNode = gTaskNodeList.GetNext(pos);
		CTaskNodeList &adjList = taskNode->adjList;
		POSITION pos2 = adjList.GetHeadPosition();
		CString cs;
		cs.Format("Processor %d period %d", taskNode->processorID, taskNode->period);
		
		while(pos2) {
			CTaskNode *taskNode2 = adjList.GetNext(pos2);
			cs += " -> ";
			CString temp2;
			temp2.Format("Processor %d period %d", taskNode2->processorID, taskNode2->period);
			cs += temp2;
		}
		cs += ";; contained components: ";
		CCompNodeList &compNodeList = taskNode->compNodeList;
		POSITION pos3 = compNodeList.GetHeadPosition();
		while(pos3) {
			CCompNode *compNode = compNodeList.GetNext(pos3);
			cs += compNode->name;
			cs += ", ";
		}
		AfxMessageBox(cs);
	}
}


void DebugDisplayCompNodeList() {
	POSITION pos = gCompNodeList.GetHeadPosition();
	while(pos) {
		CCompNode *compNode = gCompNodeList.GetNext(pos);
		CCompNodeList &adjList = compNode->adjList;
		POSITION pos2 = adjList.GetHeadPosition();
		CString cs;
		cs.Format("%s:%d:%d:%d", compNode->name, compNode->processorID, compNode->primaryPeriod,
			compNode->wcet);
		
		while(pos2) {
			CCompNode *compNode2 = adjList.GetNext(pos2);
			cs += " -> ";
			cs += compNode2->name;
		}
		AfxMessageBox(cs);
	}
}
