#include "stdafx.h"
#include "RMA.h"
#include <math.h>


void ComputeUtilizations() {
	POSITION pos = gProcessorNodeList.GetHeadPosition();
	while(pos) {
		CProcessorNode *processorNode = gProcessorNodeList.GetNext(pos);
		processorNode->utilization = 0.0;
		double maxBlocking = 0.0;
		int numTasks = 0;
		POSITION pos2 = processorNode->taskNodeList.GetHeadPosition();
		while(pos2) {
			CTaskNode *taskNode = processorNode->taskNodeList.GetNext(pos2);
			processorNode->utilization += taskNode->wcet / taskNode->period;
			numTasks++;
			if(taskNode->blockingTime / taskNode->period > maxBlocking) {
				maxBlocking = taskNode->blockingTime / taskNode->period;
			}
		}
		processorNode->utilization += maxBlocking;
		processorNode->utilizationBound = numTasks * (pow(2.0, 1.0/numTasks) - 1);
	}
}


double Inteference(CTaskNode **taskArray, int index, double responseTime, double cxtSwitchTime) {
	int j = 0;
	double sum = 0;

	for(j = 0; j < index; j++) {
		sum += ceil((taskArray[j]->releaseJitter + responseTime)/taskArray[j]->period) *
			(taskArray[j]->wcet + 2 * cxtSwitchTime);
	}
	return sum;
}

//The classic RMA response time computation. Ccs is CPU context switch time.
//ri = Ci + 2Ccs + Sum(hp(i)) (ceil((ri + Jj)/Tj) * (Cj + 2Ccs))
bool ProcessorAnalysis() {
	bool ret = true;
	POSITION pos = gProcessorNodeList.GetHeadPosition();
	while(pos) {
		CProcessorNode *pNode = gProcessorNodeList.GetNext(pos);
		int numTasks = pNode->taskNodeList.GetCount();
		CTaskNode **taskArray = new CTaskNode*[numTasks];
		POSITION pos2 = pNode->taskNodeList.GetHeadPosition();
		if(!pos2) continue;
		int i = 0;
		while(pos2) {
			taskArray[i++] = pNode->taskNodeList.GetNext(pos2);
		}

		for(i = 0; i < numTasks; i++) {
			CTaskNode *task = taskArray[i];
			double oldr, r;
			oldr = 0;
			r = task->wcet + task->blockingTime + 2 * pNode->cxtSwitchTime +
				Inteference(taskArray, i, oldr, pNode->cxtSwitchTime);
			while(fabs(oldr - r) > .01) {
				oldr = r;
				r = task->wcet + task->blockingTime + 2 * pNode->cxtSwitchTime +
					Inteference(taskArray, i, oldr, pNode->cxtSwitchTime);
				if(r> 100000000) {
					CString cs;
					cs.Format("Response time analysis for task %d on processor %d failed to converge!",
						task->id, pNode->id);
					AfxMessageBox(cs);
					//return false;
					r = 1000000.0;
				}
			}
			task->wcrt = r + task->releaseJitter;
			if(task->wcrt > task->deadline) 
				ret = false;
		}
	}
	return ret;
}

bool HolisticAnalysis() { 
	bool converged = false;
	int iter = 0;
	while(!converged) {
		//if(!ProcessorAnalysis()) return false;
		ProcessorAnalysis();
		converged = true;
		POSITION pos = gTaskNodeList.GetHeadPosition();
		while(pos) {
			CTaskNode *taskNode = gTaskNodeList.GetNext(pos);
			if(fabs(taskNode->previousWcrt - taskNode->wcrt) > 1.0) {
				converged = false;
			}
			taskNode->previousWcrt = taskNode->wcrt;
			POSITION pos2 = taskNode->adjList.GetHeadPosition();
			while(pos2) {
				CTaskNode *taskNode2 = taskNode->adjList.GetNext(pos2);
				taskNode2->releaseJitter = taskNode->wcrt;
			}
		}
		if(iter++ > 10000) {
			AfxMessageBox("Too many iterations in HolisticAnalysis!");
			break;
		}
	}
	return converged;
}

//void ComputeMaxSlowdownFactor()